#ifndef ARBREN_HPP
#define ARBREN_HPP

#include <vector>
#include "utils.PRO2"

template <class T> class ArbreNari{

  // Descripció del tipus: Arbre genèric que o bé és buit o bé
  // tot subarbre seu té el mateix nombre de subarbres
private:

  struct node_arbreNari {
    T info;
    vector<node_arbreNari*> seg;
  };

  int N;
  node_arbreNari* primer_node;

public:

  // Constructores

  ArbreNari(){
    N=0;
    primer_node=NULL;
  }

  ArbreNari(int n){
  /* Pre: n>=0 */
  /* Post: El resultat és un arbre n-ari sense cap element */
    N = n;
    primer_node= NULL;
  }
  
  ArbreNari(const T& x, int n){
    /* Pre: cert */
    /* Post: El resultat és un arbre amb x com a arrel i n fills que són arbres
       n-aris buits */
    N = n;
    primer_node= new node_arbreNari;
    primer_node->info = x;
    primer_node->seg = vector<node_arbreNari*>(N,NULL);
  }

  ArbreNari(const ArbreNari& original)
  /* Pre: cert */
  /* Post: el resultat és una còpia d'original */
  {
    N = original.N;
    primer_node = copia_node_arbreNari(original.primer_node);
  }


  // Destructora: 

  ~ArbreNari(){
    esborra_node_arbreNari(primer_node);
  }


  ArbreNari& operator=(const ArbreNari& original)
  /* Pre: cert */
  /* Post: El p.i. passa a ser una còpia d'original i qualsevol
     contingut anterior del p.i. ha estat esborrat
     (excepte si el p.i. i original ja eren el mateix objecte) */
  {
    if (this != &original) {
      esborra_node_arbreNari(primer_node);
      N = original.N;
      primer_node = copia_node_arbreNari(original.primer_node);
    }
    return *this;
  }


// Modificadores


  void a_buit()
  /* Pre: cert */
  /* Post: el p.i. és un arbre N-ari buit i qualsevol
     contingut anterior del p.i. ha estat esborrat */
  {
    esborra_node_arbreNari(primer_node);
    primer_node= NULL;
  }

  void plantar(const T& x, vector<ArbreNari>& v){
  /* Pre: La longitud d'v és el nombre de fills del paràmetre implícit */
  /* Post: l'arbre té x com a arrel i els n elements de v com
     a fills */
    node_arbreNari* aux= new node_arbreNari;
    aux->info= x;
    aux->seg = vector<node_arbreNari*>(N);
    for (int i=0; i<N; ++i) {
      aux->seg[i] = v[i].primer_node;
      v[i].primer_node = NULL;
    }
    primer_node= aux;
  }


  void fills(vector<ArbreNari> &v)
  /* Pre: el p.i. no és buit i li diem A, v és un vector buit */
  /* Post: el p.i. és buit, v passa a contenir els N fills de lâarbre A */
  {
    node_arbreNari* aux= primer_node;
    v = vector<ArbreNari> (N, ArbreNari(N));
    for (int i=0; i<N; ++i)
      v[i].primer_node = aux->seg[i];
    primer_node= NULL;
    delete aux;
  }


  //Consultores

  T arrel() const{
  /* Pre: L'arbre no és buit */
  /* Post: El resultat éss l'arrel de l'arbre */
    return primer_node->info;
  }

  bool es_buit() const{
  /* Pre: cert */
  /* Post: El resultat indica si l'arbre és buit o no */
    return (primer_node==NULL);
  }

  int aritat() const{
  /* Pre: cert */
  /* Post: El nombre de fills de l'arbre */
    return N;
  }

private:

  static node_arbreNari* copia_node_arbreNari(node_arbreNari* m){
    node_arbreNari* n;
    if (m==NULL) n=NULL;
    else {
      n = new node_arbreNari;
      n->info = m->info;
      int k = m->seg.size();
      n->seg = vector<node_arbreNari*>(k);
      for (int i=0; i<k; ++i)

	n->seg[i] = copia_node_arbreNari(m->seg[i]);
    }
    return n;
  }



  static void esborra_node_arbreNari(node_arbreNari* m){
    if (m != NULL) {
      for (int i=0; i<m->seg.size(); ++i)
	esborra_node_arbreNari(m->seg[i]);
      delete m;
    }
  }

};

#endif











